from .tensor import Tensor
from .tensor import float16,float32,float64,int16,int32,int64,bool
from .tensor import randn,zeros,ones,arange,rand,eye,zeros_like,ones_like,eye_like,rand_like,randn_like
from .tensor import uniform,tril,binomial,categorical
from .tensor import orthogonal,top_k,normal,arange_like
from .tensor import entropy
